import numpy as np

from chaco.api import AbstractPlotData, ArrayPlotData, Plot, ArrayDataSource
from traits.api import Dict, Instance, Str
from pandas import DataFrame

class PandasPlotData(AbstractPlotData):

    ''' Chaco requires a PlotData interface to manage plot/data mapping; however, pandas
    already is its own nice data handler, so this class mostly provides a wrapper that tries
    to be a hidden middle man.  As such, I try to make this PandasPlotData feel like a pandas
    dataframe.  It would be nice to have chaco communicate with a dataframe directly, but for now
    it must do so through this liason class.'''

    # The dataframe, column and row labels
    df = Instance(DataFrame)

    # Dict to store extra dataframes or series that aren't necessarily in the dataframe
    # Must pass
    extras=Dict()

    #Dictionary stores a string mapping of the data labels, which are not restricted to strings.
    #This allows users to get_data() with floats and chaco plots to access them with strings.

    # Dict mapping data series name to a mask array holding the selections
    selections = Dict()


    def __init__(self, dataframe, columnlabel='columns', indexlabel='index'):
        """PandasPlotData exposes a PlotData interface from a DataFrame.
        It is chosen that this object MUST be initialized with a dataframe.
           The intializer in ArrayPlotData will assign the names "seriesN" to
        any unlabeled selections passed into the program through the *data argument.
        Because dataframes are inherently labeled, this behavior is unnecessary 
        for basic use.

        indexlabel and columnlabel keywords let the user access the data stored in the rows/column
        labels in the dataframe when calling plot.  It is important that no columns or rows
        in the data share this name.

        Data is stored column-wise or row-wise depending on the choice of axis (0=column)."""
        
        super(AbstractPlotData, self).__init__() #AbstractPlotData has no __init__, but if that ever changes...        

        ### make traits or just leave as attributes? ###
        if len(dataframe.shape) > 2:
            raise NotImplementedError('Multidimensional dataframes of order 3 or higher \
	    are not supported by in PandasPlotData')
 
        self.columnlabel=columnlabel
        self.indexlabel=indexlabel
        
        event=self._add_dataframe(dataframe)
        self.data_changed = event        



    ### SHOULD JUST SUBCLASS ARRAY PLOT DATA
    def list_data(self, axis=0, as_strings=False):
        """ Returns a list of the names of the selections managed by this instance.  For convienence,
        axis can be 0,1 or the index label column label.
        """
        if axis==0 or axis==self.columnlabel:
            if as_strings:
                datalist=self._colmasks.keys()
            else:
                datalist=self._colmasks.values()

        elif axis==1 or axis==self.indexlabel:
            if as_strings:
                datalist=self._rowmasks.keys()
            else:
                datalist=self._rowmasks.values()

        return datalist



    def get_data(self, name):
        """ Attempts to return a name, which can be either from the dataframe, from the "extras" dict,
        or the column/index designators.  Trys them all systematically in the order dataframe, extras,
        column, row labels. """            
        try:
            return self.df[name].values
        except KeyError:
            try:
                return self.extras[name]
            except KeyError:
                try:
                    return self.df[self._colmasks[name]].values  #if name is str()
                except KeyError:
                    try:
                        self.extras[self._colmasks[name]].values
                    except KeyError: 
                        if name == self.columnlabel:
                            return self._colmasks.values()  
                        elif name == self.indexlabel:
                            return self._rowmasks.values()                                  

    def get_row_data(self, name):
        ''' Returns row data.  This is never used by plots, only a convienence method for users.'''
        return self.df.xs(name)

    def set_data(self, name, new_data):
        """ Sets the specified array as the value for either the specified
        name or a generated name.

        Implements AbstractPlotData.  
        THIS WILL ONLY SET ROW DATA IN A PANDA DATAFRAME OBJECT UNDER CURRENT SELECTION

        Parameters
        ----------
        name : string
            The name of the array whose value is to be set.
        new_data : array
            The array to set as the value of *name*.
        generate_name : Boolean
            I've eliminated this functionality for this datatype 

        Returns
        -------
        The name under which the array was set.
        """
        if not self.writable:
            return None

        if isinstance(new_data, list) or isinstance(new_data, tuple):
            new_data = np.array(new_data) #Convert to array data

        event = {}
        ### If entry is in data frame, change it
        try:
            self.df[name]=new_data
            event['changed']=[str(name)]
        except KeyError:
            self.df[self._colmasks[name]] = new_data    	    
            event['changed'] = [str(name)]  #Enforce strings, since DF names can be floats
                ### Else, add it to the "extras" array.
        self.data_changed = event       
        return name
    
    def update_dataframe(self, dataframe):
        ''' Changes dataframe in place assuming all new values are in here'''
        for column in dataframe.columns.values:
            self.set_data(column, dataframe[column])


    def set_dataframe(self, dataframe):
        ''' Completely reset plotdata with a new object.  This removes ALL 
        old data, does not cross check for overlapping keys.'''
        ### Remove old columns of dataframe ###
        removed={'removed':self._colmasks.keys()}
        ### Add new entries
        added=self._add_dataframe(dataframe)
        event=dict(added.items() + removed.items() )
        self.data_changed = event
               
    
    def _add_dataframe(self, dataframe):
        ''' Convienence function for set_dataframe to be called either by __init__
        or set_dataframe. Needed because can't evaluate self.df as a Bool() to 
        distinguish cases of initialization from overwriting dataframe.'''
        self.df=dataframe                
        self._colmasks=dict( ((str(val), val) for val in self.df.columns.values)) 
        self._rowmasks=dict( ((str(val), val) for val in self.df.index.values))
        return {'added':self._colmasks.keys()}
        

    ######## These are used by chaco to inform the plot that a certain region of data is selected

    def get_selection(self, name):
        """ Returns the selection for the given column name """
        return self.selections.get(name, None)

    def set_selection(self, name, selection):
        # Store the selection in a separate dict mapping name to its
        # selection array
        self.selections[name] = selection
        self.data_changed = True

if __name__=='__main__':
    df=DataFrame((np.random.randn(10,10)) ) 
    data=PandasPlotData(df)
    print data.df[0]
    df2=DataFrame((np.random.randn(10,10)) )
    data.set_dataframe(df2)
    print data.df[0]




''' Utilities for converting various file formats to a skspec TimeSpectra.
 To convert a list of raw files, use from_spec_files()
 To convert old-style timefile/spectral data file, use from_timefile_datafile()
 To convert spectral datafiles from Ocean Optics USB2000 and USB650, pas file list in 
 from_spec_files.
 
 Returns a skspec TimeSpectra with custom attributes "metadata", "filedict", "baseline".
 '''

__author__ = "Adam Hughes, Zhaowen Liu"
__copyright__ = "Copyright 2012, GWU Physics"
__license__ = "Free BSD"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Development"

import os

# 3RD Party Imports
from pandas import DataFrame, Series, datetime, read_csv, concat
import numpy as np

# skspec imports
from skspec.core.timespectra import TimeSpectra
from skspec.core.specindex import SpecIndex
from skspec.core.file_utils import get_files_in_dir, get_shortname


import logging
logger = logging.getLogger(__name__)


### Verified OCean Optics month naming conventions (Actually it capitalizes but this call with month.lower() ###
spec_suite_months={'jan':1, 'feb':2, 'mar':3, 'apr':4, 'may':5, 'jun':6, 'jul':7,
                   'aug':8, 'sep':9, 'oct':10, 'nov':11, 'dec':12}  

spec_dtype = np.dtype([ ('wavelength', float), ('intensity', float) ])


def get_shortname(filepath, cut_extension=False):
    ''' simply get the filename of fullpath.  Cut extension will remove file extension'''
    shortname = os.path.basename(filepath) 
    if cut_extension:
        shortname = os.path.splitext(shortname)[0]
    return shortname


### Following two functions are utilized by both data interfacing functions. ###
def extract_darkfile(filelist, return_null=True):
    ''' Attempts to pick out a dark file from a list using string matching.  Handling of darkfile
    is done by other functions.  Opted to raise errors and warnings here, instead of downstream.

    If return_null: it will return none if no dark file is found.  Otherwise it will raise an error.'''

    darkfiles=[]
    for infile in filelist:
        if "dark" in infile.lower() or "drk" in infile.lower():
            darkfiles.append(infile)

    ### If not darkfiles found, return null or raise ###       
    if len(darkfiles)==0:
        if return_null:
            return None
        else:
            raise Warning("WARNING: Darkfile not found in filelist:  \
         startfile %s - endfile %s"%(filelist[0], filelist[-1]))

    ### If multiple darkfiles found, RAISE a warning ###
    elif len(darkfiles)>1:
        raise Warning("Multiple darkfiles founds in filelist:  \
      startfile %s - endfile %s"%(filelist[0], filelist[-1]))

    else:
        return darkfiles[0]

def get_headermetadata_dataframe(dataframe, time_file_dict, name=''):
    ''' After creation of the dataframe and datetime-to-file dic, various metadata attributes 
    come together like filecount, filestart, filend etc...
    **run title becomes name of dataframe and can be adopted by plots'''
    filecount=len(dataframe.columns)
    timestart, timeend=dataframe.columns[0], dataframe.columns[-1]
    filestart, fileend=time_file_dict[timestart], time_file_dict[timeend]
    specstart, specend=dataframe.index[0], dataframe.index[-1]

    return {'filecount':filecount, 
            'timestart':timestart, 
            'timeend':timeend, 
            'filestart':filestart, 
            'fileend':fileend, 
            'specstart':specstart, 
            'specend':specend}

##########################################################
### Below are the 2 main functions to extract the data ###
##########################################################

def from_spec_files(file_list, name='', skiphead=17, skipfoot=1, check_for_overlapping_time=True, extract_dark=True):
    ''' Takes in raw files directly from Ocean optics USB2000 and USB650 spectrometers and returns a
    skspec TimeSpectra. If spectral data stored without header, can be called with skiphead=0.

    Parameters
    ----------
       name: Set name of returned TimeSpectra.
       
       check_for_overlapping_time: will raise errors if any files have identical times. Otherwise, time
                                   is overwritten.  Really only useful for testing or otherwise cornercase instances.
       
       extract_dark: Attempt to find a filename with caseinsenstive string match to "dark".  If dark spectrum
                     not found, will print warning.  If multiple darks found, will raise error.
      
       skiphead/skipfoot: Mostly for reminder that this filetype has a 17 line header and a 1 line footer.
       
    Notes
    -----
        Built to work with 2-column data only!!!

        Dataframe is constructed from a list of dictionaries.
        Each dataframe gets an appended headerdata attribute (dataframe.headerdata) which is a dictionary,
        keyed by columns and stores (infile, header, footer) data so no info is lost between files.

        Constructed to work for non-equally spaced datafiles, or non-identical data (aka wavelengths can have nans).
    '''

    dict_of_series={} #Dict of series eventually merged to dataframe   
    time_file_dict={} #Dict of time:filename (darkfile intentionally excluded)
    
    _overlap_count = 0 # Tracks if overlapping occurs

    ### If looking for a darkfile, this will find it.  Bit redundant but I'm lazy..###
    if extract_dark:
        darkfile=extract_darkfile(file_list, return_null=True)

        if darkfile:
            with open(darkfile) as f:
                header=[f.next().strip() for x in xrange(skiphead)]

            wavedata=np.genfromtxt(darkfile, dtype=spec_dtype, skip_header=skiphead, skip_footer=skipfoot) 
            darktime=_get_datetime_specsuite(header)        
            baseline=Series(wavedata['intensity'], index=wavedata['wavelength'], name=darkfile)

            file_list.remove(darkfile)
            f.close()
        else:
            baseline=None
            
    file_list = [f for f in file_list 
                 if os.path.basename(f) != '.gitignore']

    for infile in file_list:

        ###Read in only the header lines, not all the lines of the file
        ###Strips and splits in one go
        with open(infile) as f:
            header=[f.next().strip() for x in xrange(skiphead)]

        #Store wavelength, intensity data in a 2-column datatime for easy itemlookup 
        #Eg wavedata['wavelength']
        wavedata=np.genfromtxt(infile, dtype=spec_dtype, skip_header=skiphead, skip_footer=skipfoot) 

        # Extract time data from header
        datetime=_get_datetime_specsuite(header) 
        
        if datetime in time_file_dict:
            _overlap_count += 1        

        # Make sure timepoints aren't overlapping with any others
        if check_for_overlapping_time and _overlap_count:
            raise IOError('Duplicate time %s found in between files %s, %s.'
                          ' To overwrite, set check_for_overlapping_time = False.'
                           %( datetime, infile, time_file_dict[datetime] ))            
            

        time_file_dict[datetime]=infile
        dict_of_series[datetime]=Series(wavedata['intensity'], index=wavedata['wavelength'])

        f.close()

    ### Make timespec, add filenames, baseline and metadata attributes (note, DateTimeIndex auto sorts!!)
    timespec=TimeSpectra(DataFrame(dict_of_series), name=name) #Dataframe beacuse TS doesn't handle dict of series
    timespec.specunit='nm'
    timespec.filedict=time_file_dict
    timespec.baseline=baseline  #KEEP THIS AS DARK SERIES RECALL IT IS SEPARATE FROM reference OR REFERENCE..  

    ### Take metadata from first file in filelist that isn't darkfile
    for infile in file_list:
        if infile != darkfile:
            with open(infile) as f:
                header=[f.next().strip() for x in xrange(skiphead)]         
            meta_partial=_get_metadata_fromheader(header)
            break      

    meta_general=get_headermetadata_dataframe(timespec, time_file_dict) 
    meta_general.update(meta_partial)
    timespec.metadata=meta_general   

    if _overlap_count:
        logger.warn('Time duplication found in %s of %s files.  Duplicates were '
            'removed!' % (_overlap_count, len(file_list)))
            
    return timespec


def _get_datetime_specsuite(specsuiteheader):
    ''' Special, Ocean-optics specific function to get date information from a their customized header.'''
    dateline=specsuiteheader[2].split()
    year, month, day=int(dateline[6]), dateline[2], int(dateline[3]) 

    month=spec_suite_months[month.lower()]

    hrs, mins, secs=dateline[4].split(':')
    hrs=int(hrs) ; mins=int(mins) ; secs=int(secs)
    return datetime(year, month, day, hrs, mins, secs)


def _get_metadata_fromheader(specsuiteheader):
    ''' Populates metadata attributes from the speactrasuite datafile header'''
    sh=specsuiteheader #for ease in calling
    return {'dark_spec_pres':sh[4].split()[3], 'spectrometer':sh[7].split()[1],  
            'int_unit':sh[8].split()[2], 'int_time':int(sh[8].split()[3]), 
            'spec_avg':int(sh[9].split()[2]), 'boxcar':int(sh[10].split()[2]), 
            'ref_spec_pres':sh[5].split()[3], 'electric_dark_correct':sh[11].split()[4], 
            'strobe_lamp':sh[12].split()[2],'detector_nonlin_correct':sh[13].split()[4], 
            'stray_light_correct':sh[14].split()[4], 'pix_in_spec':int(sh[15].split()[6])}


######### Get dataframe from timefile / datafile #####
# Authors Zhaowen Liu/Adam Hughes, 10/15/12

def from_timefile_datafile(datafile, timefile, extract_dark=True, name=''): 
    ''' Converts old-style spectral data from GWU phys lab into  
    a dataframe with timestamp column index and wavelength row indicies.

    Creates the DataFrame from a dictionary of Series, keyed by datetime.
    **name becomes name of dataframe''' 

    tlines=open(timefile,'r').readlines()
    tlines=[line.strip().split() for line in tlines]           
    tlines.pop(0)

    time_file_dict=dict((_get_datetime_timefile(tline),tline[0]) for tline in tlines)

    ### Read in data matrix, separate first row (wavelengths) from the rest of the data
    wavedata=np.genfromtxt(datafile, dtype='float', skip_header=1)
    data, wavelengths=wavedata[:,1::], wavedata[:,0] #Separate wavelength column

    ### Sort datetimes here before assigning/removing dark spec etc...
    sorted_tfd=sorted(time_file_dict.items())
    sorted_times, sorted_files=zip(*( (((i[0]), (i[1])) for i in sorted_tfd)))

    ### Seek darkfile.  If found, take it out of dataframe. ###
    if extract_dark:
        darkfile=extract_darkfile(sorted_files, return_null=True)   

    if darkfile:    
        ####Find baseline by reverse lookup (lookup by value) and get index position

        #darkindex, darktime=[(idx, time) for idx, (time, afile) in enumerate(sorted_tfd) if afile == darkfile][0]
        darkindex=sorted_files.index(darkfile)
        darktime=sorted_times[darkindex]
        baseline=Series(data[:,darkindex], index=wavelengths, name=darkfile) 


        del time_file_dict[darktime] #Intentionally remove
        sorted_times=list(sorted_times) #Need to do in two steps
        sorted_times.remove(darktime)
        data=np.delete(data, darkindex, 1)  #Delete dark column from numpy data           
    else:
        baseline=None

    dataframe=TimeSpectra(data, columns=sorted_times, index=wavelengths)      
    

    ### Add field attributes to dataframe
    dataframe.baseline=baseline 
    dataframe.filedict=time_file_dict
    if name:
        dataframe.name=name

    ### Get headermeta data from first line in timefile that isn't darkfile.  Only checks one line
    ### Does not check for consistency
    for line in tlines:
        if line[0]==darkfile:
            pass
        else:
            meta_partial=_get_headermetadata_timefile(line[0])  #DOUBLE CHECK THIS WORKS
            break   

    ### Extract remaining metadata (file/time info) and return ###
    meta_general=get_headermetadata_dataframe(dataframe, time_file_dict) 
    meta_general.update(meta_partial)
    dataframe.metadata=meta_general
    dataframe.specunit='nm'  #This autodetected in plots    

    ### Sort dataframe by ascending time (could also sort spectral data) ###
    dataframe.sort(axis=1, inplace=True) #axis1=columns

    return dataframe

def from_gwu_chem_UVVIS(filelist, sortnames=False, shortname=True, cut_extension=False, name=''):
    ''' Format for comma delimited two column data from GWU chemistry's UVVis.  These have no useful metadata
    or dark data and so it is important that users either pass in a correctly sorted filelist.  Once the 
    dataframe is created, on can do df=df.reindex(columns=[correct order]).  
    
    It uses read_csv() to and creates a list of dataframes.  Afterwards, concat() merges these.
    
    Kwds:
       sortnames- Will attempt to autosort the filelist. Otherwise, order of files passed in is
                  directly used as columns.
       shortname- If false, full file path is used as the column name.  If true, only the filename is used. 
       
       cut_extension- If using the shortname, this will determine if the file extension is saved or cut from the data.'''

    if shortname:
        fget=lambda x:get_shortname(x, cut_extension=cut_extension)
    else:
        fget=lambda x: x
    
    ### Either full names or short names of filelist    
    working_names=[fget(afile) for afile in filelist]
        

    dflist=[read_csv(afile, sep=',', header=None, index_col=0, skiprows=2, na_values=' ',  #Used to be ' \r', or is this from IR?
                               names=[fget(afile)]) for afile in filelist]
    
    ### THIS IS BUSTED, PUTTING NANS EVERYWHERE EXCEPT ONE FILE, but dflist itself ws nice.
    dataframe=concat(dflist, axis=1)
                        
    ### concat tries to sort these, so this will preserve the sort order
    if sortnames:
        dataframe=dataframe.reindex(columns=sorted(working_names))

    dataframe=TimeSpectra(dataframe) #this is fine

    dataframe.metadata=None
    dataframe.filedict=None
    dataframe.baseline=None
    dataframe.specunit='nm' #This autodetected in plots    
    if name:
        dataframe.name=name
    
    return dataframe
            

def _get_datetime_timefile(splitline):
    ''' Takes in SPLIT line of timefile (aka a list), returns datetime object '''
    hrs, mins, secs=splitline[4].split(':')
    hrs=int(hrs) ; mins=int(mins) ; secs=int(secs)
    day, month, year=int(splitline[3]), spec_suite_months[splitline[2].lower()], int(splitline[1])

    return datetime(year, month, day, hrs, mins, secs)   

def _get_headermetadata_timefile(splitline):
    ''' Return the reduced, available components of headermetadata from timefile (missing data).''' 
    ### Take items that are present ###
    filldic=dict(zip(['int_time', 'int_unit', 'spec_avg', 'boxcar'], splitline[5:]))

    ### Append null for items that are missing ###
    missing=('spectrometer','dark_spec_pres', 'ref_spec_pres', 'electric_dark_correct', 
             'strobe_lamp', 'detector_nonlin_correct', 'stray_light_correct', 'pix_in_spec')
    missingdic=dict((item, None) for item in missing)

    ### MANUALLY ADDING SPECTROMETER FOR THIS DATATYPE! COULD BE PROBLEMATIC IF I CHANGE SPECTROMETERS
    ### OR USE THIS PROGRAM TO OPEN OTHER DATAFILES (UNLIKELY)
    missingdic['spectrometer']='USB2E7196'  

    return missingdic

def from_oceanoptics(directory, sample_by=None, sort=True, **from_spec_files_kwds):
    ''' Wrapper to from_spec_files to both generate a file list from a directory and read files into time spectra.
        
        Parameters:
        -----------
            directory: directory with raw spectral files.
            sample_by: Sample every X files.  (Useful for reducing large datsets prior to readin)
            **from_spec_files_kwds: All kwds passed to from_spec_files().
            
        Notes:
        ------
            Slice works by taking the 0 file every time and counting from there.  So if you enter sample_by=3,
            expect to get files 0, 3, 6 etc... but if sample_by_10, you get files 0,10,20
    '''
    
    files=get_files_in_dir(directory, sort=sort)

    if sample_by:    
        files=files[0::sample_by]  #Slice already takes cares of most errors
    
    return from_spec_files(files, **from_spec_files_kwds)

if __name__=='__main__':
    # Assumes datapath is ../data/gwuspecdata...
    
    ### Test of raw fiber spectral data
    
    ts=from_oceanoptics('../data/gwuspecdata/fiber1', sample_by=10, sort=True, name='oceanopticstest')

    files=get_files_in_dir('../data/gwuspecdata/fiber1')
    ts=from_spec_files(files, name='Test Spectra')
    #print ts, type(ts)
    
    ### Test of timefile/datafile
    #dfile='../data/gwuspecdata/ex_tf_df/11_data.txt'
    #tfile='../data/gwuspecdata/ex_tf_df/11_tfile.txt'
    #ts2=from_timefile_datafile(dfile, tfile, extract_dark=True, name='Test Spec 2')
    #print ts2, type(ts2)
    
    ### Test of chem UVVIS data
    files=get_files_in_dir('../data/gwuspecdata/IRData')
    ts3=from_gwu_chem_UVVIS(files, name='IR Spec')
    print ts3, type(ts3), ts3[ts3.columns[0]]
    
    ### ADD CHEM IR DATA (Did it in a notebook at one point actually)


''' Utilities in regard to custom exceptions, used throughout skspec.  Mostly used 
to alert users when they've passed incorrect attributes to a function.'''

__author__ = "Adam Hughes"
__copyright__ = "Copyright 2012, GWU Physics"
__license__ = "Free BSD"
__maintainer__ = "Adam Hughes"
__email__ = "hugadams@gwmail.gwu.edu"
__status__ = "Development"

from operator import attrgetter

class GeneralError(Exception):

    ''' Default custom error.  Basic interface to define a default error; but
        easily overwritten when the user wants to pass a custom message.'''

    default = ''

    def __init__(self, msg=None):
        self.message = msg

    def __str__(self):
        if self.message:
            return repr(self.message)
        else:
            return repr(self.default)
        
class BaselineError(GeneralError):
    
    default = 'Baseline operation failure'
    

class ParameterError(GeneralError):
    
    default = "Parmater unknown"
    
class RefError(GeneralError):
    ''' Error raised when a user-supplied iterable does not have same spectral
    values as those of the timespectra.'''
    
    def __init__(self, index, timespectra):
        sunit = timespectra.specunit
        self.message = ('Cannot resolve length %s reference (%s%s - %s%s) and length %s %s (%s%s - %s%s)'\
                      %(len(index), index[0], sunit,  index[-1], sunit, len(timespectra.reference), \
                        timespectra.name, timespectra.df.index[0], sunit, timespectra.df.index[-1], sunit) )
        
        
# See grits client to show how to make this indistinguishable from normal argparse error
class ParserError(GeneralError):
    
    default = 'Failed to construct parser'
        
class LogExit(SystemExit):
    ''' Used by logger. '''
    

# Functions 
def badvalue_error(attribute, allowedvalues):
    ''' Wrong value of argument passed.'''
    return AttributeError('Received attribute, %s, however, allowed values are:' 
    ' %s'%(attribute, allowedvalues))

def badtype_error(attribute, allowedtypes):
    ''' Wrong type of argument passed. Provide attribute and allowed types'''
    return TypeError('Received attribute "%s" of type %s; however, allowed types are %s'%(attribute, type(attribute), allowedtypes))

def badcount_error(numallowed, numreceived, numtotal, argnames=None):
    ''' Function required x of arguments y (out of z arguments total) to work, but user
        entered wrong amount.  If argnames, error will print the names of attributes required.
        Argnames should be the names of the total number of arguments (eg if numtotal = 4, argnames 
        should be 4 strings)'''

    if argnames:
        return AttributeError('requires %s of %s nonull keywords for %s but received %s.'%(numallowed, numtotal, argnames, numreceived))
    else:
        return AttributeError('requires %s of %s nonull keywords but received %s.'%(numallowed, numtotal, numreceived))


# Phase this out?
def badkey_check(key, allowedvalues, case_sensitive=False):
    ''' Either string not passed or wrong value passed.  
        allowedvalues is a list of values that the key attribute can take on. 
        Useful for case when you want to make sure the keyword has one of x values.
  
        **case_sensitive: If False, will compare attribute.lower() to allowedvalues.lower().
        
        Notes
        -----
           Messy because I added option that key can be None type and also string.  This is
           only useful so far in TimeSpectra iunits dictionary.
        
        Returns: None (Raises errors if test fails).'''
    
    ### Handle case of key=None
    if key==None:
        if None in allowedvalues:
            return
        else:
            raise badvalue_error(key, "%s"%(','.join(allowedvalues)) )
            

    ### Make sure attribute is a string
    if not isinstance(key, basestring):
        raise badtype_error(key, basestring)

    ### Can I compare lowercase strings?
    if case_sensitive ==False:
        key=key.lower()

        ### Handle case of None in possible keys (ie allowedvalues)
        if None in allowedvalues:
            allowedvalues.remove(None)        
            allowedvalues=[v.lower() for v in allowedvalues]
            allowedvalues.append(None)
        
    ### If attribute not found
    if key not in allowedvalues:
        if None in allowedvalues:
            allowedvalues.remove(None)
            allowedvalues.append('None')
        raise badvalue_error(key, "%s"%(','.join(allowedvalues)) )

    return 

### IS THIS TOO GENERIC!?  Make ones for type() and len() issues?

def null_attributes(obj, callfcn, *attr_required):
    ''' Logic gate ensures non-null values for required attributes on an object.
    
    Parameters
    ----------
       obj:  The Python object whose attributes are to be examine.
       callfcn: Reference to the function that called this, so make Errors more clear.
       attr_requried:  List of attributes that must have non-null values.
       
    Example
    -------
       If I want to make sure an object has valid values for x,y,z before continuing.
       null_attributes(obj, 'x','y','z')
       
    Notes:
    ------
    Null refers to a value of NONE, not False.
    
    The object does not have one of the required attributes to begin with, attrgetter will raise an error 
    in first line below. '''
    try:
        values=attrgetter(*attr_required)(obj)
    except AttributeError:
        raise AttributeError('%s call to null_attribute() missing some of the required attributes for examination'%callfcn)

    ### Zip together, and if only a single value (eg not iterable), just piece manually
    if len(attr_required)==1:
        dic={attr_required[0]:values}
    else:
        dic=dict(zip(attr_required, values))

    missing=[k for k in dic if dic[k] == None]
    
    if missing:
        raise AttributeError('%s requires non-null values for attribute(s): "%s".  Returned null for "%s"'%(callfcn, '","'.join(attr_required), ','.join(missing)))
    


""" Unit Class: series of metadata for various index units used throughout.
All class attributes to conserve memory.  Unit conversions are defined
in the respective index classes calling these"""


class UnitError(Exception):
   """ """

class Unit(object):
   """ Basic unit used by skspec Index objects.""" 
   short = None
   full = 'No unit' #None works, but literaly prints None 
   symbol = ''
   category = None
   
   def __init__(self, **kwargs):
      
      if 'short' in kwargs and 'full' not in kwargs or \
         'full' in kwargs and 'short' not in kwargs:
         raise UnitError('Please enter both a short and full Unit name.')
      
      for key in kwargs:
         setattr(self, key, kwargs[key])
        
   # Alternative representations for plotting/headres (not sure if useful)_
   @property
   def cat_short(self):
      """ Category(short) eg wavelength(nm)"""
      return '%s (%s)' % (self.category, self.short)

   @property         
   def cat_full(self):
      """ Category(long) eg wavelength(nanometers) """
      return '%s (%s)' % (self.category, self.full)
            
            
class IUnit(Unit):
   """ Intensity unit """
   category = 'Intensity' #Better one?
   symbol = 'I'


# Consider adding a better __repr__ to these for when objects are called?
class ConversionUnit(Unit):
   """ Used by conversion indexes to represent units that can be converted
   from one system to another (like feet to miles)
   """
   _canonical = False

   @staticmethod   
   def to_canonical(self):
      NotImplemented
      
   @staticmethod
   def from_canonical(self):
      NotImplemented
      
if __name__ == '__main__':
   unit = Unit(short='bar')
   print unit.__dict__
   print unit.short
   unit.short = 'foo'
   print unit.short


"""Provides core "TempSpectra" class and associated utilities."""

import logging
from skspec.logger import decode_lvl, logclass
from skspec.core.spectra import Spectra
from skspec.units import TEMPUNITS
from skspec.core.specindex import SpecIndex
from skspec.core.abcindex import ConversionIndex

logger = logging.getLogger(__name__) 


class TempIndex(ConversionIndex):
   """ Temperature Index """
   unitdict = TEMPUNITS  

# Ignore all class methods!
@logclass(log_name=__name__, skip = ['wraps','_dfgetattr', 'from_csv', 
                                     '_comment', '_transfer'])
class TempSpectra(Spectra):
   """Spectral Index (rows) with Temperature Variation (columns) """
   def __init__(self, *dfargs, **dfkwargs):
      dfkwargs['force_columns'] = TempIndex
      dfkwargs['force_index'] = SpecIndex
      super(TempSpectra, self).__init__(*dfargs, **dfkwargs)
      
if __name__ == '__main__':
   import numpy as np
   import matplotlib.pyplot as plt
   tempspec = TempSpectra(np.random.randn(50,50), specunit='cm-1', varunit='C')
   tempspec.plot(cbar=True)
   plt.show()


from abcunits import ConversionUnit
from math import pi


# Physical Constants/Conversion factors
H = 6.626068*10**-34          #Planck's constant m**2 kg / s
eVtoJ = 1.60217646 * 10**-19  #Number of Joules in one eV (needed becasue eV is not the standard MKS unit of energy)
C = 299792458.0               #speed of light m/s


# SPECTRAL UNITS   
class SpecUnit(ConversionUnit):
   """  Spectroscopy units (nm, m, ev etc...)
   Allow for recpricol units such as cm-1; used by converters.  All 
   conversions go through unit, considered the core unit.  Adds category
   to distinguish between wavelength, wavenumber, energy, frequency."""

   category = ''
   
   @property
   def symbol(self):
      """ Used by plotting, especially correlation analysis.   For example,
      
              label1 = r'$\bar{A}(%s_1)$' % symbol
              
      Where symbol is this.  Don't forget to wrap in r'$...$' for tex.
      """
      if self.category == 'wavelength':
         return r'\lambda'
   
      elif self.category == 'wavenumber':
         return r'\kappa'
      
      elif self.category == 'frequency':
         return r'\nu'
   
      elif self.category == 'energy':
         return r'\epsilon'
      
      elif not self.category: #NullUnit
         return ''
      
      else:

         raise UnitError('invalid category' % self.category)
   
   
class Meters(SpecUnit):
   short = 'm'
   full = 'meters'
   category = 'wavelength'
   _canonical = True
   
   @staticmethod
   def to_canonical(x):
      return x
      
   @staticmethod
   def from_canonical(x):
      return x


class Centimeters(SpecUnit):
   short = 'cm'
   full = 'centimeters'
   category = 'wavelength'

   @staticmethod   
   def to_canonical(x):
      return x / 100.0

   @staticmethod      
   def from_canonical(x):
      return 100.0 * x  

   
class Micrometers(SpecUnit):
   short = 'um'
   full = 'microns'
   category = 'wavelength'
   
   @staticmethod
   def to_canonical(x):
      return x / 100000.0 
      
   @staticmethod
   def from_canonical(x):
      return 100000.0 * x

   
class Nanometers(SpecUnit):
   short = 'nm'
   full = 'nanometers'
   category = 'wavelength'
   
   @staticmethod
   def to_canonical(x):
      return x / 1000000000.0 
      
   @staticmethod
   def from_canonical(x):
      return 1000000000.0 * x   

   
class Metersinverse(SpecUnit):
   """ Cycles per distance/wavenumber"""
   short = 'k'
   full = 'inverse meters' #or cycles per meter or radians per meter
   category = 'wavenumber'
   
   # TEST THIS
   @staticmethod
   def to_canonical(x):
      return 1.0 / x
      
   @staticmethod
   def from_canonical(x):
      return 1.0 / x

   
class Centimetersinverse(SpecUnit):
   short = 'cm-1'
   full = 'inverse centimeters'
   category = 'wavenumber'

   # DEFINE IN TERMS OF TO METERS?
   @staticmethod
   def to_canonical(x):
      return (1.0 / x) * (0.01)
      
   @staticmethod
   def from_canonical(x):
      return 1.0 / (x * 100.0)

   
class Micrometersinverse(SpecUnit):
   short = 'um-1'
   full = 'inverse microns'
   category = 'wavenumber'

   
class Nanometersinverse(SpecUnit):
   short = 'nm-1'
   full = 'inverse nanometers'
   category = 'wavenumber'

        
class Frequency(SpecUnit):
   """ lambda = v/f where v is c for electromagnetic radiation; in general,
   it's called the phase speed. """
   short = 'f'
   full = 'hertz'
   category = 'frequency'
   
   @staticmethod
   def to_canonical(x):
      meters = Meters.to_canonical(x)
      return C / meters
      
   @staticmethod
   def from_canonical(x):
      return C / x


# Verified correct, look at rad sec-1
# https://www2.chemistry.msu.edu/faculty/reusch/virttxtjml/cnvcalc.htm
class Angularfrequency(SpecUnit):
   short = 'w'
   full = 'radians per second'
   category = 'frequency'
   
   @staticmethod
   def to_canonical(x):
      meters = Meters.to_canonical(x)
      return (2.0 * pi * C) / meters
      
   @staticmethod
   def from_canonical(x):
      return (2.0 * pi * C) / x

   
class Electronvolts(SpecUnit):
   short = 'ev'
   full = 'electron volts'
   category = 'energy'

   @staticmethod
   def to_canonical(x):
      meters = Meters.to_canonical(x)
      return (H*C/(eVtoJ) ) / meters
      
   @staticmethod
   def from_canonical(x):
      return (H*C/(eVtoJ)) / x

   
_specunits = (
             Meters(), 
             Metersinverse(), 
             Nanometers(), 
             Centimeters(), 
             Centimetersinverse(),
             Micrometers(),
             Micrometersinverse(),
             Electronvolts(),
             Frequency(),
             Angularfrequency(),
             )

SPECUNITS = dict((obj.short, obj) for obj in _specunits)



""" SpecStack: Similar to pandas Panel but built to 
    store TimeSpectra classes."""

from pandas.core.panel import Panel
from skspec.core.timespectra import TimeSpectra
from skspec.core.utilities import boxcar


#class SpecStack(Panel):
    
    #_constructor_sliced = TimeSpectra    
    
    #def __init__(self, data=None, items=None, major_axis=None, minor_axis=None,
                     #copy=False, dtype=None, name=None):
        #'''Initialize with same arguments as panel, with special keywords
           #for the SpecStack object.
           
        #Parameters:
        #-----------
            #See Panel API for Panel__init__() arguments.
            #name: name of the SpecStack object.'''
            
        #super(SpecStack, self).__init__(data=data, items=items, major_axis=
                             #major_axis, minor_axis=minor_axis, copy=copy,
                             #dtype=dtype)      
       
        #self.name=name
                
    #def _get_unique(self, attr):
        #''' Inspects SpecStack itemwise for an attribute for unique values.
            #If non-unique value for the attributes are found, returns
            #"mixed". '''
        #unique=set(self.inspect(attr).values())
        #if len(unique) > 1:
            #return 'mixed'
        #else:
            #return tuple(unique)[0] #set doesn't support indexing        
        
    #def _set_attr(self, attr, val):
        #''' Set attributes itemwise.'''
        #for item in self:
            #setattr(self[item], attr, val)
            
    #def inspect(self, attr):
        #'''Returns a dictionary of items:attribute for a given attribute.'''
        #return dict((item, getattr(self[item], attr)) for item in self.items)
    
    #@property
    #def specunit(self):
        #return self._get_unique('specunit')
                
    #### Do I want to make as a _set method to avoid accidental overwrite?
    #@specunit.setter
    #def specunit(self, unit):
        #''' Sets specunit for every stored TimeSpectra.'''
        #self._set_attr('specunit', unit)
    
    #@property
    #def iunit(self):
        #return self._get_unique('iunit')    
    

    #### This shouldn't have a setter
    #@property
    #def timeunit(self):
        #return self._get_unique('timeunit')
    

    ##def apply(self, func, axis='major'):
        ##"""
        ##Apply

        ##Parameters
        ##----------
        ##func : numpy function
            ##Signature should match numpy.{sum, mean, var, std} etc.
        ##axis : {'major', 'minor', 'items'}
        ##fill_value : boolean, default True
            ##Replace NaN values with specified first

        ##Returns
        ##-------
        ##result : DataFrame or Panel
        ##"""
        ##i = self._get_axis_number(axis)
        ##result = np.apply_along_axis(func, i, self.values)
        ##return self._wrap_result(result, axis=axis)
    
    #def applynew(self, func, *args, **kwargs):
        #for item in self:
            #self[item]=getattr(self[item], func)(*args, **kwargs)

##SpecStack._setup_axes(axes      = ['items', 'major_axis', 'minor_axis'], 
                  ##info_axis = 0,
                  ##stat_axis = 1,
                  ##aliases   = { 'spectral': 'major_axis',
                                ##'temporal': 'minor_axis' },
                  ##slicers   = { 'major_axis': 'index',
                                #'minor_axis': 'columns' })

from pandas.core import panelnd
from pandas.core.generic import NDFrame
import numpy as np
from pandas import DataFrame

class PanelNew(NDFrame):

    @property
    def _constructor(self):
        return type(self)

    _constructor_sliced = DataFrame

PanelNew._setup_axes(axes      = ['items', 'spectral', 'temporal'], 
                  info_axis = 0,
                  stat_axis = 1,
                  aliases   = { 'spectral': 'major_axis',
                                'temporal': 'minor_axis' },
                  slicers   = { 'spectral': 'index',
                                'temporal': 'columns' })

if __name__=='__main__':
    ### For testing.
    #from skspec.core.specindex import SpecIndex
    #from pandas import date_range, DataFrame, Index
    #import numpy as np       

    x=TimeSpectra([1,1,1])
    y=PanelNew(x) 
    print y


""" Utilities to run notebooks offline, among others.  CREDIT FOR MOST OF THIS GOES TO:
https://github.com/ipython/ipython/wiki/Cookbook%3a-Notebook-utilities """

import os,sys,time

from Queue import Empty
try:
    from IPython.kernel import KernelManager
except ImportError:
    from IPython.zmq.blockingkernelmanager import BlockingKernelManager as KernelManager 
from IPython.nbformat.current import reads, NotebookNode
 

MAX_TIMEOUT = 60 #Seconds before cell is timed out

# IMPLEMENT LOGGING SOON

import logging
logger = logging.getLogger(__name__)


def run_nb_offline(nb_path):
    """ Read notebook from filepath and execute it; report errors in code cells.
    """
    if not os.path.isfile(nb_path):
        raise Exception('Invalid path: %s' % nb_path)
    with open(nb_path) as f:
        nb = reads(f.read(), 'json')
	
    logging.info("Running notebook %s" % nb.metadata.name)
    km = KernelManager()
    km.start_kernel(stderr=open(os.devnull, 'w'))
    try:
        kc = km.client()
    except AttributeError:
        # 0.13
        kc = km
    kc.start_channels()
    shell = kc.shell_channel
    # simple ping:
    shell.execute("pass")
    shell.get_msg()
    
    cells = 0
    failures = 0
    for ws in nb.worksheets:
        for cell in ws.cells:
            if cell.cell_type != 'code': #ONLY RUN CODE CELLS
                continue
            shell.execute(cell.input)
            # wait for finish, maximum TIMEOUT
            reply = shell.get_msg(timeout=MAX_TIMEOUT)['content']
            if reply['status'] == 'error':
                failures += 1
                logging.info("\nNotebook FAILURE:")
                logging.info(cell.input)
                logging.info('-----')
                logging.info('raised:')
                logging.info('\n'.join(reply['traceback']))
            cells += 1
#            sys.stdout.write('.')
 
    logging.info("Finished running notebook")
    logging.info( "    ran %3i cells" % cells)
    if failures:
        logging.warning( "    %3i cells raised exceptions" % failures)
    kc.stop_channels()
    km.shutdown_kernel()
    del km
 
if __name__ == '__main__':
    for ipynb in sys.argv[1:]:
        run_notebook(nb)



from pandasplotdatav2 import PandasPlotData
from traits.api import DelegatesTo, Str, Delegate, Enum

## For testing
import numpy as np
from pandas import DataFrame


### Need to pass traits into these so I can overwrite stuff at the beginning and don't have to do all this shit
class SpectralPlotData(PandasPlotData):
    ''' PandasPlotData that is built to store spectral data.  It enforces that 
    wavelength data be added to extras by default.  For aid in unde'''
    
    #Traits used for easy reference of the dataframe labels corresponding to spectral and temporal data.
    # Would prefer delegate!
    temporalaxis=Enum(0,1)  #Default to 0
    specname=Str('Wavelength')
    timename=Str('Time')
    
   
    def __init__(self, dataframe, extras=None, **kwtraits):
        ''' I choose the temporal axis to the primary one.'''
        
        self.sync_trait('temporalaxis', self, 'primaryaxis')
        if self.temporalaxis==0:
            self.sync_trait('specname', self, 'primaryname')
            self.sync_trait('timename', self, 'secondaryname')
        else:
            self.sync_trait('specname', self, 'secondaryname')
            self.sync_trait('timename', self, 'primaryname')            
                  

        super(SpectralPlotData, self).__init__(dataframe, extras=extras, \
                                            add_labels_to_extras=True,\
                                            **kwtraits) 
if __name__=='__main__':
    df=DataFrame((np.random.randn(10,10)) ) 
    data=SpectralPlotData(df, specname='HO')
    print data[0]
    df2=DataFrame((np.random.randn(10,10)) )
    data.update_dataframe(df2)
    print data.extras


from traits.api import HasTraits, Float, Array, Property, Enum, Trait, CFloat, Constant
from traitsui.api import View, Item, Group, HSplit, VSplit
from numpy import linspace, array
from math import pi

# Help text:
ViewHelp = """
This program is based on the Converter example in the TraitsUI examples archive.  This example is similar with
the following changes:
  1.  Full arrays are used in the place of single values for all the conversions.
  2.  The conversions allow for reciprocal unit systems, which is often necessary in spectrometry.  For example,
      the energy of a photon can be described in energy units, but also in units of wavelength and frequency using 
      the formulas:  E=hf  and lf=c   
	E=photon energy, c=speed of light, l=wavelength of light, f=frequency of light, c=speed of light
        Therefore, the energy and wavelength are related inversely by E=hc/l
  3.  The first and last values of the array are shown in the View to give the view some idea of the underlying array
      manipulations taking place.  This is merely for aesthetics.
"""

class SpectralConverter ( HasTraits ):
    '''Converts between many sets of spectral units'''
    h=float(6.626068*10**-34)          #Planck's constant m**2 kg / s
    eVtoJ=float(1.60217646 * 10**-19)  #Number of Joules in one eV (needed becasue eV is not the standard MKS unit of energy)
    c=float(299792458)                 #speed of light m/

    # Units trait maps all units to meters.  This allows us to use the physical constants in there MKS default units:
    Units = Trait( 'Meters', { 'Meters':      1.0,        
				   'Centimeters':.01,
				   'Micrometers':.000001,    #All of these are proportional to units of length
				   'Nanometers': .000000001,          
				   'cm-1': .01,              #All of these are inversely proportional to units of length
				   'Wavenumber(nm-1)':.000000001, 
				   'Frequency(Hz)':  c,  
				   'Angular Frequency(rad)': 2.0*pi*c,
				   'eV': h*c/(eVtoJ)  
					 } )

    #Next we specify which units are related to length directly and inversely#
    proportional=['Meters', 'Nanometers', 'Centimeters', 'Micrometers'] 
    reciprocal=['cm-1', 'eV', 'Wavenumber(nm-1)', 'Frequency(Hz)', 'Angular Frequency(rad)']             

    input_array  = Array( )   #Set on initialization
    input_units  = Units( )   #Set on initialization
    output_array = Property(Array, depends_on = [ 'input_array', 'input_units',
                                             'output_units' ])
    output_units  = Units('Frequency(Hz)')  #Default output unit

    #Property Traits#
    xstart=Property(Float, depends_on='input_array')
    xend=Property(Float, depends_on='input_array')

    xnewstart=Property(Float, depends_on='output_array')
    xnewend=Property(Float, depends_on='output_array')

    traits_view=View(  
		VSplit(
		Group(HSplit( Item('xstart', label='Old Spectral Min', style='readonly'), 
			      Item('xend', label='Old Spectral Max', style='readonly'),
			      Item('input_units' , style='simple', label='Input Units' ))  ),
		Group(HSplit( Item('xnewstart', label='New Spectral Min'),
			      Item('xnewend', label='New Spectral Max'), 
			      Item('output_units', label='Output Units')  )
		      )
		      ),  kind='modal', width=800, # help=ViewHelp,
		          buttons = [ 'Undo', 'OK', 'Cancel', 'Help' ])

    # Property implementations#
    def _get_xstart(self): return self.input_array[0]

    def _get_xend(self):  
	end=(self.input_array.shape[0])-1
	return self.input_array[end]

    def _get_xnewstart(self): return self.output_array[0]
	
    def _get_xnewend(self): 
	end=self.output_array.shape[0]-1
	return self.output_array[end]

    #This computes the new array, and depending on the relation between the input and output values,
    # it changes the output to account for the relative inversion between the quanties.  There are four
    # cases to consider:
	 #(both inputs are in proportional list--no relative inverion)
         #(both inputs are in reciprocal list--double inversion)
         #(one input in reciprocal, one in proprotional- single inversion (2 cases)
    def _get_output_array ( self ):
	if self.input_units in self.proportional and self.output_units in self.proportional:
	        return (self.input_array * self.input_units_) / self.output_units_

	elif self.input_units in self.proportional and self.output_units in self.reciprocal:
	        return 1.0/( (self.input_array * self.input_units_) / self.output_units_)

	elif self.input_units in self.reciprocal and self.output_units in self.proportional:
	        return 1.0/( (self.input_array * self.output_units_) / self.input_units_)   #Output/input

	elif self.input_units in self.reciprocal and self.output_units in self.reciprocal:
	        return  (self.input_array * self.output_units_) / self.input_units_

if __name__ == '__main__':
    #Populate an aribtrary array to mimic the visible light spectrum (400nm -700nm)#
    x=linspace(400, 700, num=100)
    f=SpectralConverter()
    f.input_array=x; f.input_units='Nanometers'
    f.configure_traits()



from operator import attrgetter
import os
import os.path as op
import shutil

import logging
logger = logging.getLogger(__name__)

# XXX: Add logging/remove print statements

def get_shortname(filepath, cut_extension=False):
    ''' simply get the filename of fullpath.  Cut extension will remove file extension'''
    shortname=os.path.basename(filepath) 
    if cut_extension:
        shortname=os.path.splitext(shortname)[0]  #Cut file extension
    return shortname

def sort_summary(summary_file, delim='\t'):
    f=open(summary_file, 'r')
    lines=f.readlines()
    lines=[line.strip().split(delim) for line in lines]
    header=lines.pop(0)
    f.close()
    ### Tries to sort by filenames then magnification of form 'f1_3000.tif'
    try:
        lines=sorted(lines, key=lambda item: (item[0].split('_')[0], int( item[0].split('_')[1].strip('.tif')) ) )
    except Exception:
        ### Following exceptions are adhoc fixes for non-standard run names.
        ## Test for style ## F1_b2_30kx... (8/13/12)
        firstname=lines[0][0].split('_')
        if 'f' in firstname[0] and 'b' in firstname[1]:
            try:
                lines=sorted(lines, key=lambda item: (item[0].split('_')[1],(item[0].split('_')[0] ) ))
            except Exception:
                print 'Could not sort summary file, failed at A'


        ## Test for style ## 8_b2_f1_... (8/6/12)                        
        elif 'f' in firstname[2] and 'b' in firstname[1]:
            try:
                lines=sorted(lines, key=lambda item: (item[0].split('_')[1],(item[0].split('_')[2] ),(item[0].split('_')[3] ) ))
            except Exception:
                print 'Could not sort summary file, failed at B'

       ### TO ADD ###     
       ## Test for style ## 8_b2_f1_... (8/6/12) (test for 822)                       

    else:
        f=open(summary_file, 'w')
        f.write(delim.join(header)+'\n\n')
        for i, line in enumerate(lines):
            ### Add a row divider between fibers
            if i > 0:
                if lines[i][0].split('_')[0] != lines[i-1][0].split('_')[0]:  #If switching fibers
                    f.write('\n')
                    
            f.write(delim.join(line) + '\n')

        f.close()

def get_files_in_dir(directory, sort=True):
    ''' Given a directory, returns just the files; ignores directories '''

    # If path entered with or without trailing '/', account for it
    directory = directory.rstrip('/')    
    files = [op.join(directory, item) for item in os.listdir(directory) if 
             op.isfile(op.join(directory, item))]

    if sort:
        files.sort()
    return files

            
def make_root_dir(rootout, overwrite=False, verbose=True):
    ''' Creates directory structure for program results output.  Warning, if overwriting, all files and subfolders
        in the root out directory will be destroyed, not just ones pertaining to the relevant infiles.'''
    ### Check if outdirectory already exists and react decide to overwrite or error ###
    if os.path.exists(rootout):
        if overwrite:
            print 'Deleting directory and all contents of, %s' %rootout
            shutil.rmtree(rootout)
        else:
            raise IOError('Directory %s already exists, remove or set overwrite to True'%rootout)
        
    ### Make output directory and subdirectories ###    
    if verbose:
        print 'making directory %s' %rootout
    os.makedirs(rootout)
    return rootout


### Deprecated functions below ###
def out_getter(imjobject, with_header=True, delim='\t'):
    ''' This is a special getter which takes in attribute fields as well as corresponding mapping functions
    to format return in a sexy way.  Couldn't quite rig it up properly so this is not in use!'''

    empty=lambda x: x #How to handle null case
    fround=lambda x: round(x, 2) #Quick rounding
  #  scaleout='%s/pixel'%units, round(scale,3) #NEED TO MAKE UNITS AN ATTRIBUTE GETTER!!!
    join_it=lambda x: delim.join([str(item) for item in x])
    
    outparms=[('shortname',empty), ('af_coverage',fround), ('bw_coverage', fround,), ('diam_min_mode_max', join_it), \
              ('percent_sampled',fround), ('scale',fround),('field_of_view', join_it,)]
    
    outfields=[item[0] for item in outparms]
    f=attrgetter(*outfields)    
    atts=f(imjobject)
    outstring=''
    if with_header:
        outstring=outstring+'#'+(delim.join(outfields))+'\n'
        
#    for i, item in enumerate(outparms):
 #       outstring=outstring+delim+str(item[1](atts[i]))
        
    outstring=outstring+delim.join(str(item[1](atts[i])) for i,item in enumerate(outparms) )+'\n'
        
    return outstring
    



float_display_units = 3 #When outputting floats to plots/printout, how many 
                        # labels to display
                        
multicols = 3 #Default number of columns when making multiplot of stacked data

PAD = 4 * ' ' #padding

#I'd like a config do diplay full or short name of time unit/spec unit/iunit on plots...

# DEFAULT PLOT Options
# --------------------

# Fonts/Labels
LABELSIZE = 'medium'
TITLESIZE = 'large'


# 1D
CMAP_1DSPECPLOT = 'coolwarm' #Divergent red/blue/gray middle

# Contour
CMAP_CONTOUR = 'seismic'  #Divergent red/blue white middle
NUM_CONTOURS = 20
FILL_CONTOUR = False


# 3D
COLOR_3DPLOT = 'gray'
PROJECTION_CMAP = 'jet' #Default color of PRoJECTION in add_projection
MESH = 10 #Strides in surf/wire plots.  If data is under this number
C_MESH = MESH
R_MESH = MESH  #If want default mesh to be non-square

# Spectra output spacing
HEADERDELIM = '\t'
HEADERHTMLDELIM = '&nbsp;' * 8

# Default specifier to Spectrum
SPECIFIERDEF = 'values' 
MISSING = '??' #When unit info is missing, header/plotting will refer to this



#Want C(abcspectra, metaframe)

from pandas import Series, DataFrame
import skspec.core.utilities as pvutils
import skspec.config as pvconfig
from pandas.core.indexing import _LocIndexer
from skspec.units.abcunits import IUnit, Unit, UnitError
import numpy as np

# Exceptions
# ----------

class SpecError(Exception):
   """ """
   
class SpecIndexError(SpecError):
   """ For indexing operations in particular."""   
   

# Primary Object
# --------------

class ABCSpectra(object):
   """ Generic methods shared by all skspec spectra objects.  INCLUDING
   SPECTRUM. 
   For example,
   printing out html orientation, and defining the "nearby" slicer.  The goal
   is to separate out methods that are intended to be customized, but might
   be buried on the already heavy Spectra object.
   """

   def __repr__(self, *args, **kwargs):
      
      header = self._header
      return ''.join(header)+'\n'+self._frame.__repr__()+'   ID: %s' % id(self)    

   def _repr_html_(self, *args, **kwargs):
      """ Ipython Notebook HTML appearance basically.  This only generates
      the colored header; under the hood, self._frame._repr_html_ calculates
      the table, including the proper size for optimal viewing and so on.
      """
      # Change output based on Spectra vs. Spectrum
      obj = self._frame
      
      # Series doesn't have _repr_html, so actually call DF's
      if isinstance(obj, Series):
         obj = DataFrame(obj, columns=[self.specifier])
   
      # Call DataFrame _repr_html
      dfhtml = obj._repr_html_(*args, **kwargs)
      return ('<h4>%s</h4>' % ''.join(self._header_html)) +'<br>'+ dfhtml
   
   @property
   def _spec_span(self):
      return pvutils._compute_span(self.index, with_unit=False)
   
   @property
   def _var_span(self):
      return pvutils._compute_span(self.columns, with_unit=False)
   
   @property
   def _intensity_span(self):
      return '%s - %s' % (self.min().min(), self.max().max())
      

   @property
   def _header(self):
      """ Header for string printout """
      delim = pvconfig.HEADERDELIM
      
      # Certain methods aren't copying this correctly.  Delete later if fix

      ftunit = pvutils.safe_lookup(self, 'full_varunit')
      spunit = pvutils.safe_lookup(self, 'full_specunit')
      iunit = pvutils.safe_lookup(self, 'full_iunit')
       
   
      header = "*%s*%s[%s X %s]%sIunit: %s\n" % \
            (self.name, 
             delim, 
             ftunit,
             spunit,
             delim, 
             iunit)
      
      return header
      
      
   @property
   def _header_html(self):
      """ Header for html printout """
      
      delim = pvconfig.HEADERHTMLDELIM
   
      if self.ndim > 1:
         colorshape = '<font color="#0000CD">(%s X %s)</font>' % (self.shape)
      else:
         colorshape = '<font color="#0000CD"> (%s)</font>' % (self.shape)

      ftunit = pvutils.safe_lookup(self, 'full_varunit')
      spunit = pvutils.safe_lookup(self, 'full_specunit')
      iunit = pvutils.safe_lookup(self, 'full_iunit')
   
      header = "%s&nbsp%s%s [%s X %s]%sIunit:&nbsp%s" % \
         (self.name, 
          colorshape,
          delim,
          ftunit,
          spunit,
          delim,
          iunit)   
      
      return header
        

   
   @property  #Make log work with this eventually?
   def full_name(self):
      """ Timespectra:name or Timespectra:unnamed.  Useful for scripts mostly """
      outname = getattr(self, 'name', 'unnamed')
      return '%s:%s' % (self.__class__.__name__, self.name)

   # Indexer
   _nearby=None

   @property
   def nearby(self, *args, **kwargs):      	
      """ Slicers similiar to loc that allows for nearby value slicing.
      It also would be posslbe to define operations on this through bool
      logic like df[(x<50) & (X>50), but that requires creating a new
      indexer]"""
      if self._nearby is None:
         try:
            self._nearby=_NearbyIndexer(self)
         # New versions of _IXIndexer require "name" attribute.
         except TypeError as TE:
            self._nearby=_NearbyIndexer(self, 'nearby')
      return self._nearby    

   @property
   def full_iunit(self):
      return self._iunit.full

   @property
   def iunit(self):
      return self._iunit.short
   
   @iunit.setter
   def iunit(self, unit):
      """ Set iunit from string or Unit object.  If from string, short
      and full are same, symbol='I'
      """
      if isinstance(unit, basestring):
         self._iunit = IUnit(short=unit, 
                            full=unit, #Short and full, same thing
                            )

      elif isinstance(unit, Unit): #Don't force IUnit type; users may not care
         self._iunit = unit 

      # Better than just iunit = None for plotting api?
      elif not unit:
         self._iunit = IUnit() #Default/null     

      else:
         raise UnitError('Unit must be a string, IUnit type or None!')


# Nearby Indexer 
# --------------
class _NearbyIndexer(_LocIndexer):
   """ Index by location, but looks for nearest values.  Warning: not all
   use cases may be handled properly; this is predominantly for range slices
   eg (df.nearby[50.0:62.4]), ie looking at a range of wavelengths.

   For more on various cases of selection by label:
   http://pandas.pydata.org/pandas-docs/stable/indexing.html#indexing-label

   Tries to handle all the cases of loc:
      A single label, e.g. 5 or 'a', (note that 5 is interpreted as a label
      of the index. This use is not an integer position along the index)

      A list or array of labels ['a', 'b', 'c']

      A slice object with labels 'a':'f' (note that contrary to usual
      python slices, both the start and the stop are included!)

      A boolean array
   """   

   def _getitem_axis(self, key, axis=0, validate_iterable=False):
      """ This is the only method that needs overwritten to preserve all
      _LocIndexer functionality.  Just need to change aspects where it
      would fail if key not found in index and replace with a key lookup.

      This is better than overwriting __getitem__ because need axis knowledge
      for slicing like [50:55, 90:95]...
      """

      labels = self.obj._get_axis(axis)
      values = labels.values #Necessary for subtraciton in find_nearest

      def _nearest(v):
         """ Find the neareast value in the Index (cast to array type);
         necessary because diff (ie subtraction) is not defined on Index
         in the way used here.
         """
         vmin, vmax = min(values), max(values)
         if v < vmin:
            raise SpecIndexError("%s is less than Index min value of %s" 
                                 % (v, vmin))
         elif v > vmax:
            raise SpecIndexError("%s is greater than Index max value of %s" 
                                 % (v, vmax))
         return values[(np.abs(values - v)).argmin()]

      if isinstance(key, slice):
         start, stop, step = key.start, key.stop, key.step
         if key.start:
            start = _nearest(key.start)
         if key.stop:
            stop = _nearest(key.stop)
         key = slice(start, stop, key.step)
         self._has_valid_type(key, axis)
         out = self._get_slice_axis(key, axis=axis)

      elif _is_bool_indexer(key):
         raise NotImplementedError('Bool indexing not supported by _Nearby Indexer')
#            return self._getbool_axis(key, axis=axis)

      elif _is_list_like(key):

         # GH 7349
         # possibly convert a list-like into a nested tuple
         # but don't convert a list-like of tuples
         # I Just chose to not support this case work
         if isinstance(labels, MultiIndex):
            raise SpecIndexError("MultiIndex nearby slicing not supported.")

         # an iterable multi-selection (eg nearby[[50, 55, 65]])
         if not (isinstance(key, tuple) and
                 isinstance(labels, MultiIndex)):

            if hasattr(key, 'ndim') and key.ndim > 1:
               raise ValueError('Cannot index with multidimensional key')

            if validate_iterable:
               self._has_valid_type(key, axis)

            # WILL OUTPUT WRONG INDEX STUFF LIKE TIMEINDEX
            # NEED TO FIX
            raise NotImplementedError("See GH #107")
            out = self._getitem_iterable(key, axis=axis)

         # nested tuple slicing
         if _is_nested_tuple(key, labels):
            locs = labels.get_locs(key)
            indexer = [ slice(None) ] * self.ndim
            indexer[axis] = locs
            out = self.obj.iloc[tuple(indexer)]

      # fall thru to straight lookup
      else:
         key = _nearest(key)  
         self._has_valid_type(key, axis)
         out = self._get_label(key, axis=axis)      

      if isinstance(out, DataFrame): #If Series or Series subclass
         out = self.obj._transfer(out) 

      # Hack becase Series aren't generated with these and 
      # self._getitem_lowerdim() will call this recursively
      elif issubclass(out.__class__, Series):
         out.nearby = self.obj.nearby
      return out



import plot_utils as put
import matplotlib.pyplot as plt
from skspec.core.utilities import split_by
from basic_plots import areaplot, range_timeplot

# Used by SpecStack.plot()
def slice_plot(ts_list, names=[], n=4, *plotargs, **plotkwds):
    """ Pass in a container of timespectra; outputs is subplot for reach. """

    default_names = ['' for i in range(len(ts_list))]
    names = put._parse_names(names, default_names)
    
    figtitle = plotkwds.pop('title', '')

    # CURRENTLY, DOES NOT TAKE AXIS
    fig, axes, plotkwds = put.multi_axes(len(ts_list), **plotkwds)
    
    for (idx, tspec) in enumerate(ts_list):
        tspec.plot(ax=axes[idx], fig=fig, title=names[idx], *plotargs, **plotkwds)
        
    fig.suptitle(figtitle, fontsize=20)        
    return axes
    

def quad_plot(ts, *plotargs, **plotkwds):
    """ Output a matplotlib figure with full spectra, absorbance, area and 
    stripchart.  Figure should be plotly convertable through py.iplot_mpl(fig)
    assuming one is signed in to plotly through py.sign_in(user, apikey).
    
    Parameters
    -----------
    title : str
        Title of the overall figure
        
    striplegend : bool (False)
        Add a legend to the strip chart
        
    colormap : string ('jet')
        Colormap applied to full and absorbance spectra.
        'Jet' is applid to strip chart regardless.
        
    tight_layout: bool (False)
        Calls mpl.fig.tight_layout()
    """

    title = plotkwds.pop('title', '')
    tight_layout = plotkwds.pop('tight_layout', False)
    figsize = plotkwds.pop('figsize', (8,8))


    f, axes = put.splot(2,2, fig=True, figsize=figsize)
    f.suptitle(title, fontsize=20)
    if tight_layout:
        f.tight_layout()
    
    cmap = plotkwds.pop('colormap', 'jet')
    strip_cmap = 'spectral'
    
    striplegend = plotkwds.pop('striplegend', False)
    
    ts.plot(*plotargs, 
             ax=axes[0], 
             title='Spectra', 
             colormap = cmap,
             fig=f, #for colorbar
             **plotkwds)


    range_timeplot(ts.wavelength_slices(8), 
                   ax=axes[1], 
                   legend=False,
                   colormap = strip_cmap,
                   title='Spectral Slices',
                   **plotkwds)    
    
    
    ts.plot(*plotargs,
             norm='r',
            ax=axes[2], 
            colormap=cmap, 
            title='Normalized',
            **plotkwds)


    areaplot(ts, *plotargs,
             ax=axes[3], 
             title='Area', 
             fig=f,
             **plotkwds)

    # Custom legend to strip chart (http://matplotlib.org/users/legend_guide.html#multicolumn-legend)
    if striplegend:
        axes[1].legend(loc='lower center',
                       ncol=4, 
                       fontsize=5, 
#                       mode='expand',
                       bbox_to_anchor=(0.5,-0.1))

 
    for a in (axes[1], axes[3]):
        a.yaxis.tick_right()      
        a.yaxis.set_label_position("right")

    # Remove y-axis of area/stripchart
    put.hide_axis(axes[0], axis='x')
    put.hide_axis(axes[1], axis='x')

    #axes[1].get_yaxis().set_ticklabels([])#set_visible(False)
    #axes[3].get_yaxis().set_ticklabels([])
    #axes[0].get_xaxis().set_ticklabels([])
    #axes[1].get_xaxis().set_ticklabels([])
    
    return f


def six_plot(ts, *plotargs, **plotkwds):
    """ Output a matplotlib figure with full spectra, absorbance, area and 
    stripchart.  Figure should be plotly convertable through py.iplot_mpl(fig)
    assuming one is signed in to plotly through py.sign_in(user, apikey).
    
    Parameters
    -----------
    title : str
        Title of the overall figure
        
    striplegend : bool (False)
        Add a legend to the strip chart
        
    colormap : string ('jet')
        Colormap applied to full and absorbance spectra.
        'Jet' is applid to strip chart regardless.
        
    tight_layout: bool (False)
        Calls mpl.fig.tight_layout()
    """

    title = plotkwds.pop('title', '')
    tight_layout = plotkwds.pop('tight_layout', False)
    figsize = plotkwds.pop('figsize', (10,8))
    
    f, axes = put.splot(3,2, fig=True, figsize=figsize)
    f.suptitle(title, fontsize=20)
    if tight_layout:
        f.tight_layout()
    
    cmap = plotkwds.pop('colormap', 'jet')
    strip_cmap = 'spectral'
    
    striplegend = plotkwds.pop('striplegend', False)
    
    ts.plot(*plotargs, 
             ax=axes[0], 
             title='Spectra', 
             colormap = cmap,
             fig=f, #for colorbar
             **plotkwds)


    range_timeplot(ts.wavelength_slices(8), 
                   ax=axes[1], 
                   legend=False,
                   colormap = strip_cmap,
                   title='Slices (Full)',
                   **plotkwds)    
    
    ts.plot(*plotargs,
            ax=axes[2], 
            colormap=cmap, 
            norm = 'r',
            title='Normalized (r)',
            **plotkwds)

    range_timeplot(ts.as_norm('r').wavelength_slices(8), *plotargs,
             ax=axes[3], 
             legend=False,
             title='Slices (r)',
             fig=f,
             **plotkwds)

    ts.plot(*plotargs,
            ax=axes[4], 
            norm='a',
            colormap=cmap, 
            title='Normalized (a)',
            **plotkwds)

    areaplot(ts, *plotargs,
             ax=axes[5], 
             title='Area', 
             fig=f,
             **plotkwds)

    # Custom legend to strip chars (http://matplotlib.org/users/legend_guide.html#multicolumn-legend)
    if striplegend:
        for ax in [axes[1]]: #axes[5] 
            ax.legend(loc='lower center',
                       ncol=4, 
                       fontsize=5, 
#                       mode='expand',
                       bbox_to_anchor=(0.5,-0.1))

    # Right axes to y-axis
    for a in (axes[1], axes[3], axes[5]):
        a.yaxis.tick_right()      
        a.yaxis.set_label_position("right")

    # Remove x-axis of area/stripchart
    for ax in (axes[0:4]):
        put.hide_axis(ax, axis='x')
    
    return f    
    
if __name__ == '__main__':
    from skspec.data import aunps_water
    import matplotlib.pyplot as plt
    ts = aunps_water()
    six_plot(ts, striplegend=True)
    plt.show()


# skspec intefrace to plotly:
# http://nbviewer.ipython.org/github/plotly/python-user-guide/blob/master/s00_homepage/s00_homepage.ipyn

import plotly.graph_objs as grobs
import numpy as np
import plot_utils as put

def make_linetrace(x, y, **tracekwargs):#, linecolor):  
    """Trace-generating function (returns a Scatter object) from timespectra"""
    # good example of trace options http://plot.ly/python/bubblecharts
    
    name = tracekwargs.pop('name', '')
    tracekwargs.setdefault('color', 'rgb(255,0,0)') #markercolor
    #tracekwargs.setdefault('opacity', 1.0) 
    #tracekwargs.setdefault('width', 1.0) 
        
    return grobs.Scatter(x=x,
                         y=y,          
                         mode='lines',          
                         name=name,          
                         marker= grobs.Line(**tracekwargs) 
                         )


def make_pointtrace(x, y, **tracekwargs):  
    """Trace-generating function (returns a Scatter object) from timespectra"""
    # good example of trace options http://plot.ly/python/bubblecharts
    
    name = tracekwargs.pop('name', '')    
    tracekwargs.setdefault('color', 'rgb(255,0,0)') #markercolor
    tracekwargs.setdefault('symbol', 'circle')
    #tracekwargs.setdefault('opacity', 1.0)
    tracekwargs.setdefault('size', 10)


    return grobs.Scatter(x=x,
                         y=y,
                         mode='markers',
                         marker=grobs.Marker(**tracekwargs)
                         )


def _parsenull(value):
    """ 1.0.22 plotly will error if gets axis titles of None (turns into null)
    in the json.  Therefore, this replaced values of None with an empty string.
    This may be fixed in later version of plotly."""
    if not value:
        value = ''
    return value


def layout(ts, *args, **kwargs):
    """ Make a plotly layout from timespectra attributes """    
    
    kwargs.setdefault('title', _parsenull(ts.name))
    kwargs.setdefault('plot_bgcolor', '#EFECEA') #gray
    kwargs.setdefault('showlegend', False)

    # Map x,y title into grobs.XAxis and grobs.YAxis
    xtitle = _parsenull(kwargs.pop('xtitle', ts.specunit))
    ytitle = _parsenull(kwargs.pop('ytitle', ts.iunit))    

   
    kwargs['xaxis'] = grobs.XAxis(title=xtitle)
    kwargs['yaxis'] = grobs.YAxis(title=ytitle)
    
    layout = grobs.Layout(**kwargs)

    axis_style = dict(zeroline=False,       # remove thick zero line
                     gridcolor='#FFFFFF',  # white grid lines
                     ticks='outside',      # draw ticks outside axes 
                     ticklen=8,            # tick length
                     tickwidth=1.5)        #   and width

    # Can I just set these in XAxis and YAxis __init__?
    layout['xaxis'].update(axis_style)
    layout['yaxis'].update(axis_style)
    return layout
    

def ply_fig(ts, points=False, color='jet', **kwds):
    """ Convert a timespectra to plotly Figure.  Figures can be then directly
    plotted with plotly.iplot(figure) in the notebook.  Use the layout keyword
    to specify layout parameters; all other keywords are interpreted as line style
    keywords (color
    
    Parameters
    ----------
    
    points: bool
        If true, scatter plot; else, lineplots
        
    color: str
        Any valid matplotlib color or colormap.
    
    layout: dict
        Dictionary of keywords that go into layout.  
        
    """
    
    data = grobs.Data()
    layoutkwds = kwds.pop('layout', {})
    lout = layout(ts, **layoutkwds)    
    
    # List of colors, either single color or color map
    try:
        cmapper = put.cmget(color) #Validate color map
    except AttributeError:
        cmapper = [color for i in range(len(ts.columns))]
    else:
        cmapper = put._df_colormapper(ts, color, axis=0)        
        
    # Map colors to rgb
    cmapper = map(put.to_normrgb, cmapper)    
    
    # Scale by 255 and string format for plotly
    def _rgbplotlycolor(rgb):
        r,g,b = rgb
        return 'rgb(%s, %s, %s)' % (255.0*r, 255.0*g, 255.0*b)
    
    cmapper = map(_rgbplotlycolor, cmapper)

    if points:
        tracefcn = make_pointtrace
    else:
        tracefcn = make_linetrace

    for idx, clabel in enumerate(ts):            
        trace = tracefcn(
            x = np.array(ts.index),               # Not necessary to force to np.array.dtype(float)
            y = np.array(ts[ts.columns[idx]]),
            name=clabel,
            color=cmapper[idx], #marker color
            **kwds
            ) 
        data.append(trace)

    return grobs.Figure(data=data, layout=lout)


#def ply_multifig(*figargs):
    #""" """
    #for fig in figargs:
        #ts,  
    
    
if __name__ == '__main__':
    from skspec.data import test_spectra
    ts = test_spectra()
    out = ply_figure(ts)

    print tls.get_subplots(rows=3, columns=2, print_grid=True)

    
    import plotly.plotly as py
    py.sign_in('reeveslab', 'hfpi35mejw')
    
    fig = ply_figure(ts, color='black', points=True)
    py.iplot(fig, filename='foo', fileopt='new')    
    print 'FINISHED PLOT 1'



#
# Canopy product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is confidential and NOT open source.  Do not distribute.
#

''' Client logging tools.  configure_logger() returns a root_logger with separate
    file and stream handlers.  log() is a flexible decorator that logs a function,
    displays its arguments and their values, and formatts errors different based
    on the value of logging.trace; logging.testmode.
    
    In the future, this should be consolidated into a Class'''

import inspect       
import logging        
import sys

from traceback import print_exception
from functools import wraps

from skspec.exceptions import LogExit


logger = logging.getLogger(__name__) 

# STRING FORMATTING
def _fmt_word(word, max_chars = 60, crop_front=True):
    '''str-convert a word.  If it exceeds max_chars, crop it to max_chars - 3
       and prefix it with "..."  eg:
          areallylongword --->  ....lylongword
        
    '''        
    word = str(word)
    if len(word) > max_chars:
        if crop_front:
            word='...' + word[-(max_chars-3):]
        else:
            word=word[:(max_chars -3)] + '...'
    return word

def configure_logger(screenlog=True, logfile=None, screen_level=logging.INFO,
                file_level=logging.DEBUG, time=True, name=None, **filekws):
    ''' Instantiates a logger with streamhandler, filehandler or both.
        Very similar to logging.basicConfig except it allows for simultaneous
        to-screen and to-file output.

        Parameters
        ----------
        screenlog : bool
            Log handler writes to screen.

        logfile : str (filename)
            Filename to which loghandler writes

        screen_level : str of logging.CATEGORY or int
             Set logging level for screen handler (error, info, debug etc...)

        file_level : str of logging.CATEGORY or int
             Set logging level for file handler (error, info, debug etc...)

        time : bool
            Prefix log messages with date and time IE:
               08-12 10:45:27 DEBUG    path/to/module "message"
                            vs.
               DEBUG   path/to/module "message"


        Notes
        -----
            Apparently one must set the root logger level to info, regardless
            of the child handler level settings to get everything working.  To
            limit the handler to just ONE type of error, it then requires the
            use of Filters.

            The show_all keyword merely changes the root logger name between
            __name__ and '' which gives this behavior.

        Returns
        -------
        logger : logging.Logger instance

       '''

    # Formatting for message, date, time
    out_fmt = '%(levelname)-8s %(name)-12s:  %(message)s'

    # Ascii fmt of time, if time is enabled
    asci_fmt = '%m-%d %H:%M:%S'

    # Output time with logs?
    if time:
        out_fmt = '%(asctime)s ' + out_fmt

    if not screenlog and not logfile:
        raise ValueError('Cannot create logger without stream.')
    
    # If users pass string args to screen_level, file_level
    screen_level= decode_lvl(screen_level)
    file_level = decode_lvl(file_level)

    root_logger = logging.getLogger('')
    root_logger.setLevel(logging.DEBUG)  # Needs to be set; despite handler levels
    
    # Reset any currnet handlers (useful if multiple files call this)
    root_logger.handlers = []
    
    # Add file handler to logger
    if logfile:        

        file_handler = logging.FileHandler(logfile, **filekws)
        file_handler.setFormatter(logging.Formatter(out_fmt, asci_fmt))
        file_handler.setLevel(file_level)
        root_logger.addHandler(file_handler)

    # Add screen handler to logger
    if screenlog:        

        screen_handler = logging.StreamHandler()
        screen_handler.setFormatter(logging.Formatter(out_fmt, asci_fmt))
        screen_handler.setLevel(screen_level)
        root_logger.addHandler(screen_handler)

    # Better here than getLogger('') so that its root status is retained
    root_logger.name = name
    root_logger.debug('Initialized root logger')
    
    return root_logger


#XXX This needs tested more, particularly when wrapping fcn w/ kwargs
def log(level='info', show_args=False, show_values=False, crop_values=True,
        log_name=None, method=False):
    '''  Decorates a function and logs its call signature.

    Parameters
    ----------
    level : str/int 
        valid logging level (eg 'info', 'error', 40) at which to log mesage.
        
    show_args : bool
        If true, function arguments names will be displayed.  For example:
            foo('arg1', 'arg2', 'arg3')
            
    show_values : bool
        If true, function argument values will be display.  For example:
            foo( 30, <FooObject at 320323>, {'bar':'baz'}
            
    If show_args and show_values, call signature is reconstructed fully:
            foo ( 'arg1' = 30, 'arg2' = <FooObject at 320323> ...)
            
    crop_values : bool
        If true, values longer than certain character limit ( utils._fmt_word() )
        will be truncated to end with "...".  This is useful when function 
        arguments are long lists of values.
        
    log_name : str or None
        This decorator does not attempt to inspect the function it is wrapping,
        so log_name is used to explicitly pass the name which will go to getLogger().
      
    Notes
    -----
        description
        
    Returns
    -------
    attr : type
        description
      
    '''

    # Configure logger if it hasn't been done already
    if not logging.getLogger().handlers:
        screen_level='warning'
        if '-debug' in sys.argv:
            screen_level='debug'
        elif '-info' in sys.argv:
            screen_level='info'
        configure_logger(screen_level=screen_level)
         
    
    # Either use module logger or (@ so can tell where decorator called)
    if log_name:
        logger = logging.getLogger(log_name)   
    else:
        logger = logging.getLogger(__name__)
        
    def _wraps(fcn_name, fcn, *args, **kwargs):

        arg_names = fcn.func_code.co_varnames    

        if show_args:
            arg_str = arg_names
        
        if show_values:
            val_str = args #kwargs?
            if crop_values:
                val_str = [_fmt_word(val, crop_front=False) for val in val_str]
            
        if show_args and show_values:
                
            # This is failing for functions with *args/**kwargs
            try:
                outstr = ', '.join([(str(arg_names[i]) +' = ' + str(val_str[i]) ) \
                                for i in range(len(args))])
            except IndexError:
                logger.debug('LOG WRAPPER FAILED FOR %s' % fcn_name)
                outstr = arg_str  
            else:  
                outstr = '(' + outstr + ')'
        
        elif show_args:
            outstr = arg_str
            
        elif show_values:
            outstr = val_str
        
        else:
            outstr = '()'
        
        # Log command ( *@* to distinguish from manually logged commands)
        logger.log(decode_lvl(level), '%s%s' % (fcn_name, outstr))
        
        try:            
            
            return fcn(*args, **kwargs)             
        
        except Exception as exc:
            
            ex_type, ex, tb = sys.exc_info()
        
            # -t USES sys.argv (until fixed; argparse must use -t)
            trace = False
            if '-t'  or '--trace' in sys.argv:
                trace = True
            
            if trace:
                trc_msg = '{See trace below}'
            else:
                trc_msg = '{-t, --trace for full traceback}'            
        
            fmt_str = '' if trace else '%s: %s' % (ex.__class__.__name__, exc)
            logger.error('%s() failed %s %s\n', fcn_name, fmt_str, trc_msg)                
        
            if trace:
                print_exception(ex_type, ex, tb)
            
            raise LogExit 
    

    def deco(fcn):
        ''' Generic logger that writes function name and arugment namesto INFO. 
            It also wraps a class method, if the "method" keyword was passed
            to log().  Its up to the user to make sure actual methods are passed
            when setting method=True.'''
        
        if method:
            
            # Wrap a method, add "self" to args and call function
            def wraps(self, *args, **kwargs):
                           
                fcn_name = getattr(self, 'name', '') +'.' + fcn.__name__
                args = list(args)
                args.insert(0, self)
                return _wraps(fcn_name, fcn, *args, **kwargs)
                    
        else:
            def wraps(*args, **kwargs):
                fcn_name = fcn.__name__
                return _wraps(fcn_name, fcn, *args, **kwargs)
            
                       
        return wraps
    return deco

        
# These logging facilities ought to be consolidated into a class.        
def decode_lvl(level):
    ''' Converts logging level code ('info, 'debug') to int for logging.log()'''

    if isinstance(level, basestring):
        try:
            level = int(level)
        except ValueError:       
            level = getattr(logging, level.upper())

    return level

def logclass(public_lvl='info', private_lvl='debug', skip=[], **logkwargs):
    ''' Adds a log decorator to all class methods.  Private methods (_foo)
        are logged to debug.  Public methods are logged to info.  Magic methods
        (__foo__) are skipped; as well as methods in the "skip" list.
        
        Note: This is a class method, so instance attributes like "name"
              aren't defined.  Individual methods should add more explicit
              logging for this reason.'''

    def wraps(cls):
        logkwargs['method'] = True
        
        magic = [method for method in dir(cls) if method.startswith('__') and \
                 method.endswith('__')]
                    
        for method in dir(cls):
            value = getattr(cls, method)
    
            # skip attributes, magic methods
            if not callable(value) or method in magic or method in skip:
                continue 
            else:
                if method.startswith('_'):
                    setattr(cls, method, log(private_lvl, **logkwargs)(value))  
                else:
                    setattr(cls, method, log(public_lvl, **logkwargs)(value)) 
        return cls
    return wraps


""" Define various notebook styles that may be launched easily in-line in an 
active notebook.  Same IO API as pyparty.data"""

import os.path as op
from skspec import bundled_dir
from IPython.display import display, HTML

def load(path):
    with open((op.join(bundled_dir, path)), 'r') as f:
        return f.read()

def gwu():
    """ Ivory background; Computer Modern Roman for text """
    return load('gwu.css')

def plain():
    return ''


__all__ = {
           'gwu': gwu(),
           'plain' : plain()
           }

# In-line function called in the notebook to update the style.  This is the
# work of Brian Granger; all credit to him for this and his help in general
# discussions of notebooks styling.
# https://github.com/ellisonbg/talk-cplug2013/blob/master/load_style.py
def load_style(s, figsize=None, loghide=False):
    """Load a CSS stylesheet in the notebook either by builtin styles, or
    from a file, or from a URL.

    Examples::
        %load_style 
        %load_style mystyle.css
        %load_style http://ipynbstyles.com/otherstyle.css
    """
    if s in __all__:
        style = __all__[s]
    
    elif s.startswith('http'):
        try:
            import requests
        except ImportError:
            raise ImportError('Failed to import python "requests" library;'
                              'please install to use load_style(url)')
        r =requests.get(s)
        style = r.text

    else:
        try:
            with open(s, 'r') as f:
                style = f.read()
        except IOError:
            raise IOError('Failed to load style as a url, file or builtin type.'
               ' Valid builtins are "%s"' % '","'.join(__all__.keys() ))

    out = '<style>\n{style}\n</style>'.format(style=style)
    
    if figsize:
        if figsize == True:
            figsize = 8, 5.5
        else:
            try:
                fx, fy = figsize
            except Exception:
                fx, fy = figsize, figsize
            
        fstring = "\ninput:pylab.rcParams['figure.figsize'] = %s, %s\n" \
            % (fx,fy)
        out += fstring

    display(HTML(out))
    
    if loghide:
        import warnings #supress non-skspec log msgs
        warnings.filterwarnings('ignore')             
    
def load_ipython_extension(ip):
    """Load the extension in IPython."""
    ip.register_magic_function(load_style)


''' Provides composition class, MetaPandasObject, which is an ordinary python object that stores a Dataframe and 
attempts to promote attributes and methods to the instance level (eg self.x instead of self.df.x).  This object
can be subclassed and ensures persistence of custom attributes.  The goal of this MetaPandasObject is to provide a 
subclassing api beyond monkey patching (which currently fails in persisting attributes upon most method returns 
and upon derialization.'''

from types import MethodType
import copy
import functools
import cPickle
import collections

from pandas.core.indexing import _IXIndexer, _iLocIndexer, _LocIndexer

from pandas import DataFrame, Series

# for testing
from numpy.random import randn

# Logging functions
from skspec.logger import logclass
import logging
logger = logging.getLogger(__name__) 

#----------------------------------------------------------------------
# Store attributes/methods of dataframe for later inspection with __setattr__
# Note: This is preferred to a storing individual instances of self._frame with custom 
#       attr as if user tried self.a and self._frame.a existed, it would call this...
_dfattrs=[x for x in dir(DataFrame) if '__' not in x]

#----------------------------------------------------------------------
# Loading (perhaps change name?) ... Doesn't work correctly as instance methods

def mload(inname):
    ''' Load MetaPandasObject from file'''
    if isinstance(inname, basestring):
        inname=open(inname, 'r')
    return cPickle.load(inname)

def mloads(string):
    ''' Load a MetaPandasObject from string stored in memory.'''
    return cPickle.loads(string)        
        

# Log all public/private methods to debug
#@logclass(public_lvl='debug', log_name=__name__, skip=['_transfer'])
class MetaPandasObject(object):
    ''' Base composition for subclassing pandas DataFrame and Series.'''
    
    cousin = None #OVERWRITE
    
    def __init__(self, *dfargs, **dfkwargs):
        ''' Stores a dataframe under reserved attribute name, self._frame'''      
        self._frame = self.cousin(*dfargs, **dfkwargs)

    ### Save methods    
    def save(self, outname):
        ''' Takes in str or opened file and saves. cPickle.dump wrapper.'''
        if isinstance(outname, basestring):
            outname=open(outname, 'w')
        cPickle.dump(self, outname)

    def dumps(self):
        ''' Output TimeSpectra into a pickled string in memory.'''
        return cPickle.dumps(self)    

    #----------------------------------------------------------------------
    # Overwrite Dataframe methods and operators

    def __getitem__(self, keyslice):
        ''' Item lookup.  If output is an interable, _transfer is called.  
        Sometimes __getitem__ returns a float (indexing a series) at which 
        point we just want to return that.'''

        frameout=self._frame.__getitem__(keyslice)

        try:
            iter(frameout)  #Test if iterable without forcing user to have collections package.
        except TypeError:
            return frameout
        else:
            return self._transfer(self._frame.__getitem__(keyslice) )               

    def __setitem__(self, key, value):
        self._frame.__setitem__(key, value)    

    ### These tell python to ignore __getattr__ when pickling; hence, treat this like a normal class    
    def __getstate__(self): return self.__dict__
    def __setstate__(self, d): self.__dict__.update(d)    

    def __getattr__(self, attr, *fcnargs, **fcnkwargs):
        ''' Tells python how to handle all attributes that are not found.
        Basic attributes are directly referenced to self._frame; however, 
        instance methods (like df.corr() ) are handled specially using a
        special private parsing method, _framegetattr().'''

        # Return basic attribute        
        try:
            refout = getattr(self._frame, attr)
        except AttributeError:
            raise AttributeError('Either failed to find attribute "%s" in %s or '
             'its underlying pandas object, or it is found, but throwing an error'
             ' when accessed.'%(attr, self.__class__.__name__))           
           
        if not isinstance(refout, MethodType):
            return refout

        # Handle instance methods using _framegetattr().
        # see http://stackoverflow.com/questions/3434938/python-allowing-methods-not-specifically-defined-to-be-called-ala-getattr
        else:         
            return functools.partial(self._framegetattr, attr, *fcnargs, **fcnkwargs)
            # This is a reference to the fuction (aka a wrapper) not the function itself
            
    def __setattr__(self, name, value):
        ''' When user sets an attribute, this tries to intercept any name conflicts.  For example, if user attempts to set
        self.columns=50, this will actually try self._frame.columns=50, which throws an error.  The behavior is acheived by
        using dir() on the data frame created upon initialization, filtering __x__ type methods.   Not guaranteed to work 100%
        of the time due to implicit possible issues with dir() and inspection in Python.  Best practice is for users to avoid name
        conflicts when possible.'''
        
        super(MetaPandasObject, self).__setattr__(name, value)        
        if name in _dfattrs:
            setattr(self._frame, name, value)
        else:
            self.__dict__[name]=value


    def _transfer(self, dfnew):
        """ Copy current attributes into a new dataframe.  For methods that
        return a dataframe and need to append current attributes/columns/index.
        """
        newobj = copy.deepcopy(self)  #This looks like None, but is it type (MetaPandasObject, just __union__ prints None
        newobj._frame = dfnew
        
        # THESE ARE NEVER TRANSFERED AT DF LEVEL, JUST CREATED NEW.  TRY
        # df.loc
        # a = df*50
        # a._loc  ---> Will be None
        #newobj._loc = self._loc
        #newobj._iloc = self._iloc
        #newobj._ix = self._ix         
        return newobj


    def deepcopy(self):
        ''' Make a deepcopy of self, including the dataframe.'''
        return copy.deepcopy(self)  

    def _framegetattr(self, attr, *fcnargs, **fcnkwargs):
        ''' Called by __getattr__ as a wrapper, this private method is used 
        to ensure that output is MetaPandasObject.  For example, MetaDataFrame
        will return DataFrame, MetaSeries will return Series.  This is done
        through inspecting the output object type and comparing with self.cousin.

        **kwargs: use_base - If true, program attempts to call attribute on
        the reference.  reference ought to be maintained as a series, and 
        Series/Dataframe API's must be same.

        *fcnargs and **fcnkwargs are passed to the dataframe method.

        Note: tried to ad an as_new keyword to do this operation in place, 
        but doing self=dfout instead of return dfout didn't work.  Could 
        try to add this at the __getattr__ level; however, may not be worth it.
        '''

        out=getattr(self._frame, attr)(*fcnargs, **fcnkwargs)

        ### If operation returns a dataframe, return new TimeSpectra
        if isinstance(out, self.cousin): #metadataframe or won't have _transfer method
            dfout=self._transfer(out)
            return dfout

        ### Otherwise return whatever the method return would be
        else:
            return out

    def _repr_html_(self):
        """ Allows ipython notebook to display as default html"""
        return self._frame._repr_html_()

    def __repr__(self):
        return self._frame.__repr__()
    
    def _fast_xs(self, *args, **kwargs):
        print 'in fastxs'
        return self._frame._fast_xs(*args, **kwargs)  

    ### Operator overloading ####
    ### In place operations need to overwrite self._frame
    def __add__(self, x):
        return self._transfer(self._frame.__add__(x))

    def __sub__(self, x):
        return self._transfer(self._frame.__sub__(x))

    def __mul__(self, x):
        return self._transfer(self._frame.__mul__(x))

    def __div__(self, x):
        return self._transfer(self._frame.__div__(x))

    def __truediv__(self, x):
        return self._transfer(self._frame.__truediv__(x))

    ### From what I can tell, __pos__(), __abs__() builtin to df, just __neg__()    
    def __neg__(self):  
        return self._transfer(self._frame.__neg__() )

    ### Object comparison operators
    def __lt__(self, x):
        return self._transfer(self._frame.__lt__(x))

    def __le__(self, x):
        return self._transfer(self._frame.__le__(x))

    def __eq__(self, x):
        return self._transfer(self._frame.__eq__(x))

    def __ne__(self, x):
        return self._transfer(self._frame.__ne__(x))

    def __ge__(self, x):
        return self._transfer(self._frame.__ge__(x))

    def __gt__(self, x):
        return self._transfer(self._frame.__gt__(x))     

    def __len__(self):
        return self._frame.__len__()

    def __nonzero__(self):
        return self._frame.__nonzero__()

    def __contains__(self, x):
        return self._frame.__contains__(x)

    def __iter__(self):
        return self._frame.__iter__()
    
    def __pow__(self, exp):
        return self._transfer(self._frame.__pow__(exp))
    
    def __abs__(self):
        return self._transfer(self._frame.__abs__())     
    
    # Reflected operands (http://www.rafekettler.com/magicmethods.html)
    def __radd__(self, x):
        return self._transfer(self._frame.__radd__(x))

    def __rsub__(self, x):
        return self._transfer(self._frame.__rsub__(x))
    
    def __rmul__(self, x):
        return self._transfer(self._frame.__rmul__(x))    

    def __rdiv__(self, x):
        return self._transfer(self._frame.__rdiv__(x))
    
    def __rtruediv__(self, x):
        return self._transfer(self._frame.__rtruediv__(x))    
    
    def __rmod__(self, x):
        return self._transfer(self._frame.__rmod__(x))
    
    def __rdivmod__(self, x):
        return self._transfer(self._frame.__rdivmod__(x))    
    
    def subtract(self, *args, **kwargs):
        return self._transfer(self._frame.subtract(*args, **kwargs))

    @property
    def data(self):
        """ Accesses self._frame.  RETURNS COPY SO USER DOESNT OVERWRITE IN PLACE"""
        return self._frame.copy(deep=True)    

    @property
    def index(self):
        return self._frame.index
    
    #To avoid accidentally setting index ie ts.index()
    @index.setter
    def index(self, index):
        self._frame.index = index
        

    ## Fancy indexing
    _ix=None     
    _iloc=None
    _loc=None

        
    @property	  	
    def ix(self, *args, **kwargs):      	
        ''' Pandas Indexing.  Note, this has been modified to ensure that series returns (eg ix[3])
        still maintain attributes.  To remove this behavior, replace the following:
        
        self._ix = _MetaIXIndexer(self, _IXIndexer(self) ) --> self._ix=_IXIndexer(self)
        
        The above works because slicing preserved attributes because the _IXIndexer is a python object 
        subclass.'''
        if self._ix is None:
            try:
                self._ix=_IXIndexer(self)
            #New versions of _IXIndexer require "name" attribute.
            except TypeError as TE:
                self._ix=_IXIndexer(self, 'ix')
        return self._ix   

    @property	  	
    def iloc(self, *args, **kwargs):      	
        """ See pandas.Index.iloc; preserves metadata"""
        if self._iloc is None:
            try:
                self._iloc =_iLocIndexer(self)
            #New versions of _IXIndexer require "name" attribute.
            except TypeError as TE:
                self._iloc=_iLocIndexer(self, 'iloc')
        return self._iloc   
    

    @property	  	
    def loc(self, *args, **kwargs):      	
        """See pandas.Index.loc; preserves metadata"""
        if self._loc is None:
            try:
                self._loc = _LocIndexer(self)
            #New versions of _IXIndexer require "name" attribute.
            except TypeError as TE:
                self._loc= _LocIndexer(self, 'loc')
        return self._loc         
    
            


#@logclass(public_lvl='debug', log_name=__name__, skip=['_transfer'])
class MetaDataFrame(MetaPandasObject):

    # Repeated for clarity
    cousin = DataFrame

    @property
    def columns(self):
        return self._frame.columns

    @columns.setter
    def columns(self, columns):
        self._frame.columns = columns

        
#@logclass(public_lvl='debug', log_name=__name__, skip=['_transfer'])
class MetaSeries(MetaPandasObject):
    
    cousin = Series
        
	            
    

# Test Subclass
# -------------
class SubFoo(MetaDataFrame):
    ''' Shows an example of how to subclass MetaPandasObject with custom attributes, a and b.'''

    def __init__(self, a, b, *dfargs, **dfkwargs):
        self.a = a
        self.b = b    

        super(SubFoo, self).__init__(*dfargs, **dfkwargs)

    def __repr__(self):
        return "Hi I'm SubFoo. I'm not really a DataFrame, but I quack like one."


#### TESTING ###
if __name__ == '__main__':

    ### Create a MetaPandasObject
    meta_df=MetaDataFrame(abs(randn(3,3)), index=['A','B','C'], columns=['c11','c22', 'c33'])    
    
    meta_df.to_csv('deletejunkme')

    ### Add some new attributes
    meta_df.a=50
    meta_df.b='Pamela'
    print 'See the original metadataframe\n'
    print meta_df
    print '\nI can operate on it (+ - / *) and call dataframe methods like rank()'
    
    meta_df.ix[0]

    ### Perform some intrinsic DF operations
    new=meta_df*50.0
    new=new.rank()
    print '\nSee modified dataframe:\n'
    print new

    ### Verify attribute persistence
    print '\nAttributes a = %s and b = %s will persist when new metadataframes are returned.'%(new.a, new.b)

    ### Demonstrate subclassing by invoking SubFoo class
    print '\nI can subclass a dataframe an overwrite its __repr__() or more carefully __bytes__()/__unicode__() method(s)\n'
    subclass=SubFoo(50, 200, abs(randn(3,3)), index=['A','B','C'], columns=['c11','c22', 'c33'])    
    print subclass
    ### Access underlying dataframe
    print '\nMy underlying dataframe is stored in the "data" attribute.\n'
    print subclass.data

    ### Pickle
    print '\nSave me by using x.save() / x.dumps() and load using mload(x) / mloads(x).'
    
    # INDEX SLICING THROUGH LOC AND ILOC
    print '\nslicing one column \n'
    print meta_df.iloc[0]
    print meta_df._frame.iloc[0]

    print '\nslicing many columns\n'
    print  meta_df.iloc[0:1]
    print meta_df._frame.iloc[0:1]
    
    print '\nIndexing by label\n'
    print meta_df.loc['A':'B']

    #GOOD EXPLANATION OF WHY CANT STRING SLICE COLUMNS
    #http://stackoverflow.com/questions/11285613/selecting-columns    
    print '\nColumn slicing\n'
    print meta_df[['c11','c33']]
    print meta_df[['c11','c33']].a
    
#    df.save('outpath')
#    f=open('outpath', 'r')
#    df2=load(f)    

    print '\nCreating MetaSeries'
    s = MetaSeries(range(0,10))
    print s, type(s)
    print '\nSeries Slicing 0:5'
    print s[0:5]



from skspec.core.spectra import Spectra
from skspec.units.abcunits import Unit

class AnyFrame(Spectra):
    """ Spectra with default non-strict columns or index
    """

    def __init__(self, *dfargs, **dfkwargs):
        dfkwargs.setdefault('strict_columns', None)   
        dfkwargs.setdefault('strict_index', None)    
        
        
        dfkwargs.setdefault('specunit', Unit())    
        dfkwargs.setdefault('varunit', Unit())    
        
        
        super(AnyFrame, self).__init__(*dfargs, **dfkwargs)  
        
        
if __name__ == '__main__':
    import numpy as np
    af=AnyFrame(np.random.rand(50,50))
    af.specunit = Unit(short='foo', full='bar')
    print af
